/*
 *	SunTroff -  A program to display the output of Device Independent
 *		Troff as a window on a Sun Workstation.
 *
 *	Troff_parse.c - Parse the device independent troff file.  All
 *		the real work of drawing the bits is done by other
 *		files.
 *
 *	Authors - The original version of this program was written by
 *			Richard L. Hyde (Purdue)
 *			David Slattengren (Berkeley)
 *		It was nearly all rewritten, cleaned up and a more elegant
 *		user interface installed by
 *			Malcolm Slaney (Schlumberger Palo Alto Research)
 *		Combine characters into words
 *			David Cahlander (Cray Research, Inc.)
 *
 *	Legalese -  This command was developed as an independent project
 *		to satisfy a need of the author.  This program may contain
 *		bugs and the user is cautioned to independently verify that
 *		the program is suitable for the user's intended purpose.
 *		The program is made available on an ``as is'' basis with
 *		all faults and without any implied or expressed warranties
 *		or support from either the author, Malcolm Slaney, or
 *		Schlumberger Palo Alto Research Laboratory.
 *
 *		I am putting this program in the public domain.  You are
 *		free to use it as you wish.  In return I ask two things.
 *		First, that you do not remove the names of the authors
 *		from this work.  Secondly, if you make changes or
 *		improvements to this program that you pass these back to
 *		the author so that everybody can benefit from the
 *		improvements.
 *
 *					Malcolm Slaney  (December 1986)
 *					Schlumberger Palo Alto Research
 *					3340 Hillview Avenue
 *					Palo Alto, CA 94304
 *					(415) 496-4669
 *					spar!malcolm@decwrl.dec.com
 *					malcolm@ecn.purdue.edu
 *					malcolm@spar.slb.com (Someday)
 */

#include	"suntroff.h"

static int StopSeen = 0;
int	TroffType = T_DITROFF;	/* can be changed in app-defaults */

extern char *GetWord(), *GetLine();

#define ONLY_BERKELEY(x) \
	if (TroffType == T_BERKELEY) \
		; \
	else { \
		fatal("%s command only supported for berkeley trofftype", \
			(x)); \
		break; \
	}

#define ONLY_GROFF(x) \
	if (TroffType == T_GROFF) \
		; \
	else \
		fatal("%s command only supported for berkeley trofftype", \
			(x))

ParseInput(pn)
int pn;
{
	int	n, k, c, npn, inword = 0, h, hadd = 0;
	char	*s, ch;
	char	Buffer[BUFSIZ];

	StopSeen = 0;

	if (debug(SEEKDEBUG))
		printf("ParseInput(%d) called\n", pn);
	if (TroffType == T_BERKELEY) SetSavedFontPosition(pn);
	for (;;) {
		if ((c = GetChar()) != EOF) {
			switch (c) {
			case '\n':	
				break;
			case ' ':	/* when input is text */
			case 0:		/* occasional noise creeps in */
				break;
			case '{':	/* push down current environment */
				push_env();
				break;
			case '}':
				pop_env();
				break;
					/* two motion digits plus a
					   character */
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				HorizontalMove((c-'0')*10 + GetChar()-'0');
				PutCharacter(GetChar());
				break;
			case 'c':	/* single ascii character */
				PutCharacterString();
				break;
			case 'C':
				GetWord(Buffer, BUFSIZ);
				PutSpecialCharacter(Buffer);
				break;
			case 'u':	/* groff kerned text */
				ONLY_GROFF("u");
				hadd = GetNumber();
				(void) GetChar();	/* space */
				/* FALLTHROUGH */
			case 't':	/* groff/berkeley straight text */
				if (TroffType == T_DITROFF) {
					fatal("t command only supported for berkeley or groff trofftypes");
				}

				h = 0;
				inword = 1;
				if (TroffType == T_GROFF) {
					while((ch = GetChar()) != EOF) {
						if (ch == ' ' || ch == '\n')
							break;
						h = PutCharacter(ch);
						HorizontalMove(h + hadd);
					}
					hadd = 0;
					break;
				}
				
				/* TroffType == T_BERKELEY */
				if (GetLine(Buffer,BUFSIZ) == NULL){
					fatal("unexpected end of input");
				}
				s = Buffer;

				while ((ch = *s++) != '\n') {
					if (ch == '\\') {
						char str[3];
						
						switch (ch = *s++) {
						case '\\':
						case 'e':
						    h = PutCharacter('\\');
						    break;
						case '(':
						    str[0] = *s++;
						    str[1] = *s++;
						    str[2] = '\0';
						    PutSpecialCharacter(str);
						    h = 0;
						    break;
						}
					} else
					    h = PutCharacter(ch);
					HorizontalMove(h);
				}
				break;
			case 'D':	/* draw function */
				GetLine(Buffer, BUFSIZ);
				ParseDrawFunction(Buffer);
				break;
			case 's':	/* ignore fractional sizes */
				n = GetNumber();
				if (TroffType == T_GROFF)
					/* convert to Device dots */
					n = n * 72 / DeviceResolution;
				SetFontSize(n);
				break;
			case 'f':
				n = GetNumber();
				SetFont(n);
				break;
			case 'i':
				ONLY_BERKELEY("i");
				n = GetNumber();
				setstip(n);	/* ignore for now */
				break;
			case 'H':	/* absolute horizontal motion */
				k = GetNumber();

				if (inword == 1)
					break;
				HorizontalGoto(k);
				break;
			case 'h':	/* relative horizontal motion */
				k = GetNumber();
				HorizontalMove(k);
				break;
			case 'w':	/* word space */
				inword = 0;
				c = GetChar();
				if (c < '0' || c > '9') {
					UnGetChar(c);
					break;
				}
				HorizontalMove((c-'0')*10 + GetChar()-'0');
				PutCharacterString();
				break;
			case 'V':
				inword = 0;
				n = GetNumber();
				VerticalGoto(n);
				break;
			case 'v':
				inword = 0;
				n = GetNumber();
				VerticalMove(n);
				break;
			case 'P':	/* new spread */
				break;
			case 'p':	/* new page */
				inword = 0;
				UnGetChar(c);
				npn = RememberPagePosition();
				if (npn != pn)
					return(npn);
				(void) GetChar();
				(void) GetNumber();
				break;
			case 'n':	/* end of line */
				inword = 0;
				GetNumber();
				GetNumber();
				HorizontalGoto(0);
				break;
			case '#':	/* comment */
				GetLine((char *) NULL, BUFSIZ);
				break;
			case 'x':	/* device control */
				ParseDeviceControl();
				break;
			default:
				fprintf(stderr,
					"Line %d, Unknown input character %c(%d)\n",
					LineNumber, c, c);
				break;
			}
		} else {
			extern int LastPage, CurrentPage;
			if (!LastPage && !StopSeen){
				warning("File is incomplete!\nEnd of file reached before finding\nthe end of the document.\nHave read %d pages.",CurrentPage);
			}
			return(CurrentPage);
		}
	}
}

struct	state {
	int	size;
	int	font;
	int	style;
	int 	thick;
	int	h_pos;
	int	v_pos;
};

struct	state 	StateStack[MAXSTATE];
struct	state	*statep = StateStack;
	  
push_env()					/* begin a new block */
{
	statep->size = cursize;
	statep->font = curfont;
	statep->style = linmod;
	statep->thick = linethickness;
	statep->h_pos = HorizontalPosition;
	statep->v_pos = VerticalPosition;
	if (statep+1 >= StateStack+MAXSTATE)
		warning( "{ nested too deep");
	else
		statep++;
}

pop_env()					/* pop to previous state */
{
	if (--statep < StateStack)
		warning("extra }");
	else {
		cursize = statep->size;
		curfont = statep->font;
	    	HorizontalPosition = statep->h_pos;
		VerticalPosition = statep->v_pos;
		linmod = statep->style;
		linethickness = statep->thick;
		SetFont(curfont);
		SetFontSize(cursize);
	}
}

static char *
cons(c1, c2)
{
	static char cmd[3];

	cmd[0] = c1;
	cmd[1] = c2;
	cmd[2] = '\0';
	return cmd;
}

ParseDrawFunction(buf)
char	*buf;
{
	int	n, m, n1, m1;
	int bordered; /* until we do polygons right */

	switch (buf[0]) {
	case 'l':				/* draw a line */
		sscanf(buf+1, "%d %d", &n, &m);
		DrawLine(n, m);
		break;
	case 'c':				/* circle */
		sscanf(buf+1, "%d", &n);
		DrawCircle(n);
		break;
	case 'e':				/* ellipse */
		sscanf(buf+1, "%d %d", &m, &n);
		DrawEllipse(m, n);
		break;
	case 'a':				/* arc */
		sscanf(buf+1, "%d %d %d %d", &n, &m, &n1, &m1);
		DrawArc(n, m, n1, m1);
		break;
	case '~':				/* wiggly line */
		DrawSpline(buf+1,1);
		break;
	case 'q':				/* versatec polygon - ignore */
		ONLY_BERKELEY(cons('D', buf[0]));
		while (buf[strlen(buf) - 1] != '\n'){
			if (GetLine(buf, sizeof(buf)) == NULL){
				fatal("unexpected end of input");
			}
		}
		break;
	case 'P':				/* unbordered */
		ONLY_BERKELEY(cons('D', buf[0]));
		bordered = 0;
	case 'p':				/* polygon */
		ONLY_BERKELEY(cons('D', buf[0]));
		sscanf(buf+1, "%d", &n);
		n = 1;
		while(buf[n++] == ' ');
		while(isascii(buf[n]) && isdigit(buf[n])) n++;
		DrawSpline(buf+n, -1);
		bordered = 1;
		break;
	case 'g':				/* gremlin spline */
		ONLY_BERKELEY(cons('D', buf[0]));
		DrawSpline(buf+1, 0);
		break;
	case 'z':				/* Bezier curve */
		ONLY_BERKELEY(cons('D', buf[0]));
		warning("substituting B-spline for Bezier curve %s", buf);
		DrawSpline(buf+1,1);
		break;
	case 't':				/* line thickness */
		ONLY_BERKELEY(cons('D', buf[0]));
		sscanf(buf+1, "%d", &n);
		drawthick(n); 
		break;
	case 's':	/* line style */
		ONLY_BERKELEY(cons('D', buf[0]));
		sscanf(buf+1, "%d", &n);
		drawstyle(n);
		break;
	default:
		warning("unknown drawing function %s", buf);
		break;
	}
} 

ParseDeviceControl()				/* Parse the x commands */
{
        char str[20], str1[50];
	int c, n;
	extern int LastPage, CurrentPage;

	GetWord(str, 20);
	switch (str[0]) {			/* crude for now */
	case 'T':				/* output device */
		GetWord(DeviceName, 10);
		break;
	case 'i':				/* initialize */
		UserTypesetter = LoadDevice(DeviceName);
		CurrentPage = 1;
		SetPrinter(DeviceName);
		InitTypesetter();
		break;
	case 't':				/* trailer */
		break;
	case 'p':				/* pause -- can restart */
		break;
	case 's':				/* stop */
		StopSeen = 1;
		LastPage = CurrentPage;
		return;
	case 'r':				/* resolution when prepared */
		DeviceResolution = GetNumber();
		break;
	case 'f':				/* font used */
		n = GetNumber();
		(void) GetWord(str, 20);
		(void) GetLine(str1, 50);
		SetFontPosition(n, str, str1);
		break;
	case 'H':				/* char height */
		break;
	case 'S':				/* slant */
		break;
	}
	while ((c = GetChar()) != '\n')		/* skip rest of input line */
		if (c == EOF)
			return;
	return;
}
